require 'csv'

module GeolocationService::Services
  class ImportBulkDataService < BaseService
    option :file_path
    option :contracts, default: proc { DEFAULT_CONTRACTS }

    Contract = GeolocationService::Contracts::BulkDataContract
    Structs = GeolocationService::Structs

    DEFAULT_CONTRACTS = {
      ip: GeolocationService::Contracts::IpAddressContract.new,
      location: GeolocationService::Contracts::LocationContract.new,
      city: GeolocationService::Contracts::CityContract.new,
      country: GeolocationService::Contracts::CountryContract.new
    }.freeze

    def call
      start_time = Time.zone.now
      load_existing_records
      load_new_records

      CSV.foreach(file_path, headers: :first_row) do |row|
        ip = build_or_find_ip(row)
        if ip.nil?
          @invalid_records += 1
          next
        end

        city = find_record(:city, row['city']&.downcase)
        if city.nil?
          country = find_or_build_country(row)
          city = build_city(row, country)
        end

        build_location(row, ip, city)
      end

      bulk_save(@new_records[:ip].values, GeolocationService::Structs::IpStruct)
      bulk_save(@new_records[:country].values, GeolocationService::Structs::CountryStruct)
      bulk_save(@new_records[:city].values, GeolocationService::Structs::CityStruct)
      bulk_save(@new_records[:location], GeolocationService::Structs::LocationStruct)

      GeolocationService::ImportResult.new(
        imported_records: {
          ip: @new_records[:ip].values.size,
          city: @new_records[:city].values.size,
          country: @new_records[:country].values.size,
          location: @new_records[:location].size
        },
        invalid_records: @invalid_records,
        time_consumed: (Time.zone.now - start_time)
      )
    end

    private

    def build_or_find_ip(row)
      return if row['ip_address'].blank?

      existing_ip = find_record(:ip, row['ip_address'])
      return existing_ip if existing_ip.present?

      if validate(:ip, address: row['ip_address'], mystery_value: row['mystery_value']).success?
        new_ip = Structs::IpStruct.new(@ip_count, row['ip_address'], row['mystery_value'])
        @new_records[:ip][row['ip_address']] = new_ip
        @ip_count += 1
        return new_ip
      end

      nil
    end

    def build_city(row, country)
      return if row['city'].blank?

      if validate(:city, name: row['city'], country_id: country&.id).success?
        new_city = Structs::CityStruct.new(@city_count, row['city'], country&.id)
        @new_records[:city][row['city'].downcase] = new_city
        @city_count += 1
        return new_city
      end

      nil
    end

    def find_or_build_country(row)
      return if row['country_code'].blank?

      existing_country = find_record(:country, row['country_code'])
      return existing_country if existing_country.present?

      if validate(:country, name: row['country'], code: row['country_code']).success?
        new_country = Structs::CountryStruct.new(@country_count, row['country_code'], row['country'])
        @new_records[:country][row['country_code']] = new_country
        @country_count += 1
        return new_country
      end

      nil
    end

    def build_location(row, ip, city)
      if validate(:location, latitude: row['latitude']&.to_d, longitude: row['longitude']&.to_d, ip_id: ip&.id,
        city_id: city&.id).success?
        @new_records[:location] << Structs::LocationStruct.new(ip.id, row['latitude']&.to_d, row['longitude']&.to_d,
          city&.id)
      end
    end

    def load_existing_records
      @existing_records = {
        ip: Hash[Ip.all.collect { |ip| [ip.address, ip] }],
        city: Hash[City.all.collect { |city| [city.name.downcase, city] }],
        country: Hash[Country.all.collect { |country| [country.code.downcase, country] }]
      }
    end

    def load_new_records
      @ip_count = Ip.count == 0 ? 0 : Ip.last.id + 1
      @city_count = City.count == 0 ? 0 : City.last.id + 1
      @country_count = Country.count == 0 ? 0 : Country.last.id + 1
      @invalid_records = 0
      @new_records = {ip: {}, location: [], city: {}, country: {}}
    end

    def find_record(type, key)
      return if key.nil?

      @existing_records[type][key] || @new_records[type][key]
    end

    def validate(type, **args)
      contracts[type].call(args)
    end

    def bulk_save(all_records, struct, chunk: 1000)
      all_records.each_slice(chunk) do |records|
        ActiveRecord::Base.connection.execute(struct.insert_sql(records))
      end
    end
  end
end
